#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase.exception
      @file: exceptions.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
from hspylib.core.exception.exceptions import HSBaseException


class FirebaseAuthenticationError(HSBaseException):
    """Raised when authenticate to Firebase"""


class InvalidFirebaseCredentials(HSBaseException):
    """Raised when invalid credentials are provided to Firebase"""


class UnprocessableFile(HSBaseException):
    """Raised when invalid or binary file fails to load"""


class FirebaseException(HSBaseException):
    """Raised when Firebase module raises an error"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase
      @file: __main__.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""

from clitt.core.tui.tui_application import TUIApplication
from firebase.__classpath__ import classpath
from firebase.core.firebase import Firebase
from hspylib.core.enums.charset import Charset
from hspylib.core.tools.commons import syserr
from hspylib.core.zoned_datetime import now
from hspylib.modules.application.exit_status import ExitStatus
from hspylib.modules.application.version import Version
from textwrap import dedent

import logging as log
import os
import sys
import urllib3

# Disable this warning because we are hitting our own database
urllib3.disable_warnings()


class Main(TUIApplication):
    """Firebase Agent - Manage your firebase integration."""

    # The welcome message
    DESCRIPTION = classpath.get_source("welcome.txt").read_text(encoding=Charset.UTF_8.val)

    # location of the .version file
    VERSION = Version.load(load_dir=classpath.source_path)

    # The resources folder
    RESOURCE_DIR = classpath.resource_path

    def __init__(self, app_name: str):
        super().__init__(app_name, self.VERSION, self.DESCRIPTION.format(self.VERSION), resource_dir=self.RESOURCE_DIR)
        self.firebase = Firebase()

    def _setup_arguments(self) -> None:
        # fmt: off
        self._with_options() \
            .option(
                'config-dir', 'd', 'config-dir',
                "the configuration directory. If omitted, the User's home will be used.",
                nargs='?', default=os.getenv('HOME', self._root_dir))
        self._with_chained_args('operation', 'the Firebase operation to process') \
            .argument('setup', 'setup your Firebase account') \
            .argument('upload', 'upload files to your Firebase Realtime Database') \
                .add_option(
                    'dest-dir', 'd', 'dest-dir',
                    'the destination directory. If omitted, the current directory will be used.',
                    nargs='?', default=self._root_dir) \
                .add_option(
                        'glob', 'g', 'glob',
                        'filter path names matching a specified glob pattern.',
                        nargs='?') \
                .add_parameter('db_alias', 'alias to identify the firebase object to fetch') \
                    .add_parameter('files', 'list of files to upload (separated by a space)', nargs='*') \
            .argument('download', 'download files from your Firebase Realtime Database') \
                .add_option(
                    'dest-dir', 'd', 'dest-dir',
                    'the destination directory. If omitted, the current directory will be used.',
                    nargs='?', default=self._root_dir) \
                .add_parameter('db_alias', 'alias to identify the firebase object to fetch')
        # fmt: on

    def _main(self, *params, **kwargs) -> ExitStatus:
        """Run the application with the command line arguments"""
        log.info(
            dedent(
                f"""
        {self._app_name} v{self._app_version}
        Settings ==============================
                STARTED: {now("%Y-%m-%d %H:%M:%S")}
        """
            )
        )
        return self._exec_application()

    def _exec_application(self) -> ExitStatus:
        """Execute the specified firebase operation"""
        op = self.get_arg("operation")
        if op == "setup" or not self.firebase.is_configured():
            self.firebase.setup()
        # Already handled above
        if op == "setup":
            log.debug("Operation is setup but it was already handled")
        elif op == "upload":
            self.firebase.upload(self.get_arg("db_alias"), self.get_arg("files"), self.get_arg("glob"))
        elif op == "download":
            self.firebase.download(self.get_arg("db_alias"), self.get_arg("dest-dir"))
        else:
            syserr(f"### Unhandled operation: {op}")
            self.usage(ExitStatus.FAILED)
        return ExitStatus.SUCCESS


if __name__ == "__main__":
    # Application entry point
    Main("firebase").INSTANCE.run(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase.domain
      @file: firebase_dto.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
from hspylib.core.enums.charset import Charset
from hspylib.core.metaclass.classpath import AnyPath
from hspylib.core.zoned_datetime import now
from hspylib.modules.security.security import b64_decode, b64_encode
from typing import List

import json
import logging as log
import os

from firebase.exception.exceptions import UnprocessableFile


class FirebaseDto:
    """Represents a Firebase DTO."""

    @staticmethod
    def from_file(filepath: AnyPath, contents: str | bytes, encoding: str = Charset.UTF_8.val) -> "FirebaseDto":
        """Create a new file entry with the specified contents and expected size.
        :param filepath: the file path.
        :param contents: the file content.
        :param encoding: the file encoding type.
        """
        entry = FirebaseDto(
            str(filepath), len(contents), str(contents, encoding=encoding) if isinstance(contents, bytes) else contents
        )

        return entry

    @staticmethod
    def from_json(json_string: str) -> List["FirebaseDto"]:
        """Convert a JSON formatted string into a Firebase DTO.
        :param json_string: the JSON string to be converted.
        """
        json_obj = json.loads(json_string)
        return list(map(lambda o: FirebaseDto(o["path"], o["size"], o["data"], o["modified"]), json_obj))

    def __init__(self, file_path: AnyPath, size: int = 0, data: str = None, modified: str = None):
        self.path = str(file_path)
        self.size = size
        self.data = data
        self.modified = modified or now()

    def __str__(self) -> str:
        return str({"path": self.path, "data": str(self.data), "size": self.size, "modified": self.modified})

    def load(self) -> "FirebaseDto":
        """Loads the file contents."""
        if os.path.exists(self.path):
            try:
                with open(self.path, "r", encoding=Charset.UTF_8.val) as f_in:
                    self.data = f_in.read()
                    if (size := len(self.data)) == 0:
                        log.warning('Nothing to be loaded. File "%s" is empty', self.path)
                    self.size = size
                    return self
            except UnicodeDecodeError as err:
                raise UnprocessableFile(f'Unable to load file "{self.path}". Is it binary?') from err
        raise FileNotFoundError(f'File "{self.path}" could not be loaded because it does not exist!')

    def save(self) -> "FirebaseDto":
        """Saves current DTO content (overwrites current file content)."""
        if os.path.exists(self.path):
            log.warning('File "%s" exists and will be overwritten', self.path)
        with open(self.path, "w+", encoding=Charset.UTF_8.val) as f_out:
            f_out.write(self.data)
            f_out.flush()
            return self

    def encode(self) -> "FirebaseDto":
        """B64-Encode the entry contents."""
        self.data = b64_encode(self.data)
        return self

    def decode(self) -> "FirebaseDto":
        """B64-Decode the entry contents."""
        self.data = b64_decode(self.data)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase.core
      @file: firebase.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""

from firebase.core.agent_config import AgentConfig
from firebase.core.file_processor import FileProcessor
from firebase.core.firebase_auth import FirebaseAuth
from hspylib.core.preconditions import check_argument
from typing import List

import logging as log
import os


class Firebase:
    """Provides the firebase application functionalities."""

    def __init__(self) -> None:
        self.processor = FileProcessor()
        filename = os.environ.get(
            "HHS_FIREBASE_CONFIG_FILE", f"{os.environ.get('HOME', os.curdir)}/firebase.properties"
        )
        self.agent_config = AgentConfig(filename)

    def __str__(self):
        return str(self.agent_config)

    def setup(self) -> None:
        """Setup a firebase configuration by creating or reading an existing configuration file."""
        status = False
        while status is not None and not status:
            status = self.agent_config.prompt()
        log.debug("New firebase setup: %s", self.agent_config)

    def upload(self, db_alias: str, file_paths: List[str], glob_exp: str) -> bool:
        """Upload files to firebase.
        :param db_alias: the firebase database alias.
        :param file_paths: the file paths to be uploaded. File paths will be filtered by glob expressions.
        :param glob_exp: the GLOB expressions to filter the input files/folders.
        """
        self._authenticate()
        url = f"{self.agent_config.url(db_alias)}.json"
        if len(file_paths) <= 0:
            log.warning("Skipping zero-sized files..")
        log.debug("Uploading files  alias=%s  files=[%s]", db_alias, ",".join(file_paths))
        return self.processor.upload_files(url, file_paths, glob_exp) > 0

    def download(self, db_alias: str, dest_dir: str) -> bool:
        """Download specified aliased files from firebase.
        :param db_alias: the database alias to download from.
        :param dest_dir: the destination directory.
        """
        self._authenticate()
        url = f"{self.agent_config.url(db_alias)}.json"
        log.debug("Downloading files  alias=%s  dest_dir=%s", db_alias, dest_dir)
        return self.processor.download_files(url, dest_dir or os.environ.get("HOME")) > 0

    def is_configured(self) -> bool:
        """Checks whether firebase is properly configured or not."""
        return self.agent_config is not None and self.agent_config.firebase_configs is not None

    def _authenticate(self) -> None:
        """Authenticate to Firebase using the user UID."""
        configs = self.agent_config.firebase_configs
        FirebaseAuth.authenticate(configs.project_id, configs.uid)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase.core
      @file: agent_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
from clitt.core.tui.minput.input_validator import InputValidator
from clitt.core.tui.minput.minput import MenuInput, minput
from collections import defaultdict
from datasource.firebase.firebase_configuration import FirebaseConfiguration
from firebase.core.firebase_auth import FirebaseAuth
from firebase.exception.exceptions import FirebaseAuthenticationError
from hspylib.core.config.app_config import AppConfigs
from hspylib.core.config.properties import Properties
from hspylib.core.enums.charset import Charset
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.commons import dirname, file_is_not_empty, syserr, sysout, touch_file
from hspylib.core.tools.dict_tools import get_or_default_by_key
from hspylib.modules.cli.keyboard import Keyboard
from os.path import basename
from typing import Any, Optional

import logging as log
import os


class AgentConfig(metaclass=Singleton):
    """Holds the firebase agent configurations."""

    @staticmethod
    def wait_keystroke(wait_message: str = "%YELLOW%%EOL%Press any key to continue%EOL%%NC%") -> None:
        """Wait for a keypress (blocking).
        :param wait_message: the message to present to the user.
        """
        sysout(wait_message)
        Keyboard.wait_keystroke()

    def __init__(self, filename: str) -> None:
        self.app_configs = AppConfigs(dirname(filename), basename(filename))
        self.firebase_configs = None
        if file_is_not_empty(self.filename):
            self._load()
            log.debug("Found and loaded a Firebase configuration: %s", self)

    def __str__(self) -> str:
        return str(self.firebase_configs)

    def __repr__(self):
        return str(self)

    def __getitem__(self, item) -> Any:
        return self.app_configs[item]

    def prompt(self) -> Optional[bool]:
        """Create a new firebase configuration by prompting the user for information."""
        config = None
        if os.path.exists(self.filename):
            config = Properties.read_properties(self.filename)
        else:
            touch_file(self.filename)
        config = config.as_dict if config else defaultdict()
        title = (
            "%ORANGE%### Firebase Setup ###%EOL%"
            f"{'-' * 30} %EOL%"
            "Please fill in your Realtime Database configurations"
        )
        # fmt: off
        form_fields = MenuInput.builder() \
            .field() \
                .label('UID') \
                .validator(InputValidator.custom(r'[a-zA-Z0-9\-]')) \
                .min_max_length(28, 28) \
                .dest("uid") \
                .value(get_or_default_by_key(config, 'UID', '')) \
                .build() \
            .field() \
                .label('PROJECT_ID') \
                .validator(InputValidator.custom(r'[a-zA-Z0-9\-\_]')) \
                .min_max_length(8, 50) \
                .dest("project_id") \
                .value(get_or_default_by_key(config, 'PROJECT_ID', '')) \
                .build() \
            .field() \
                .label('EMAIL') \
                .validator(InputValidator.anything()) \
                .min_max_length(14, 50) \
                .dest("email") \
                .value(get_or_default_by_key(config, 'EMAIL', '')) \
                .build() \
            .field() \
                .label('DATABASE') \
                .validator(InputValidator.custom(r'[a-zA-Z0-9\-\_]')) \
                .min_max_length(4, 50) \
                .dest("database") \
                .value(get_or_default_by_key(config, 'DATABASE', '')) \
                .build() \
            .build()
        # fmt: on
        if result := minput(form_fields, title):
            config["UID"] = result.uid
            config["PROJECT_ID"] = result.project_id
            config["EMAIL"] = result.email
            config["DATABASE"] = result.database
            return self._setup(dirname(self.filename), basename(self.filename), config)

        return None

    def _setup(self, load_dir: str, filename: str, config_dict: dict) -> bool:
        """Setup firebase from a dict configuration
        :param load_dir: the directory where to load the configurations from.
        :param filename: the configuration file name.
        :param config_dict: firebase configuration dictionary.
        """
        try:
            sysout("%BLUE%Authenticating to Firebase ...%NC%")
            if user := FirebaseAuth.authenticate(config_dict["PROJECT_ID"], config_dict["UID"]):
                if user.uid != config_dict["UID"]:
                    raise FirebaseAuthenticationError(f"Provided UID: {config_dict['UID']} is invalid!")
                config_dict["UID"] = user.uid
                self.firebase_configs = FirebaseConfiguration.INSTANCE or FirebaseConfiguration.of(
                    load_dir, filename, config_dict
                )
                self.firebase_configs.update(dict(config_dict))
                self._save()
                sysout("%GREEN%Firebase authentication succeeded!%EOL%")
                return True
        except FirebaseAuthenticationError as err:
            syserr(err)
            self.wait_keystroke()

        return False

    @property
    def filename(self) -> str:
        """Return the configuration file."""
        file = self.app_configs["hhs.firebase.config.file"]
        return file if file else f"{os.getenv('HOME', os.getcwd())}/.firebase"

    @property
    def project_id(self) -> Optional[str]:
        """Return the firebase Project ID."""
        return self.firebase_configs.project_id

    @property
    def uid(self) -> Optional[str]:
        """Return the firebase User ID."""
        return self.firebase_configs.uid

    @property
    def email(self) -> Optional[str]:
        """Return the firebase user's email."""
        return self.firebase_configs.email

    @property
    def database(self) -> Optional[str]:
        """Return the firebase project database name."""
        return self.firebase_configs.database

    def url(self, db_alias: str) -> str:
        """Return the firebase project URL."""
        final_alias = db_alias.replace(".", "/")
        return self.firebase_configs.url(final_alias)

    def _load(self) -> None:
        """Load configurations from file."""
        self.firebase_configs = FirebaseConfiguration.of_file(self.filename)

    def _save(self) -> None:
        """Save current firebase configurations to file."""
        with open(self.filename, "w+", encoding=Charset.UTF_8.val) as f_config:
            f_config.write(str(self))
            sysout(f"%GREEN%Firebase configuration saved => {self.filename} !%NC%")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase.core
      @file: firebase_auth.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
from abc import ABC
from firebase.exception.exceptions import FirebaseAuthenticationError, InvalidFirebaseCredentials
from firebase_admin import auth, credentials
from firebase_admin.auth import UserNotFoundError, UserRecord
from firebase_admin.exceptions import FirebaseError
from hspylib.core.preconditions import check_not_none
from typing import Optional

import firebase_admin
import os


class FirebaseAuth(ABC):
    """Firebase authentication utils
    Ref: https://www.youtube.com/watch?v=esqNgnayVE8
    """

    APP = None

    @staticmethod
    def _credentials(project_id: str) -> credentials.Certificate:
        """Create Firebase credentials based on the configured credentials file.
        :param project_id: the Firebase Realtime database project ID.
        """
        creds_file = check_not_none(
            os.environ.get("HHS_FIREBASE_CREDS_FILE", f"{os.environ.get('HOME')}/firebase-credentials.json"), project_id
        )
        try:
            creds = credentials.Certificate(creds_file.format(project_id=project_id))
        except (IOError, KeyError, ValueError) as err:
            raise InvalidFirebaseCredentials(f'Invalid credentials or credential file "{creds_file}"') from err

        return creds

    @classmethod
    def authenticate(cls, project_id: str, uid: str) -> Optional[UserRecord]:
        """Authenticate to Firebase using valid credentials.
        :param project_id: the Firebase Realtime database project ID.
        :param uid: the Firebase User ID.
        """
        try:
            cls.APP = firebase_admin.initialize_app(FirebaseAuth._credentials(project_id)) if not cls.APP else cls.APP
            if user := auth.get_user(uid):
                return user
            raise FirebaseAuthenticationError(f"Failed to authenticate to Firebase. User ID '{uid}' not found.")
        except UserNotFoundError as err:
            raise FirebaseAuthenticationError(f"Failed to authenticate to Firebase => {err}") from err
        except (ValueError, FirebaseError) as err:
            raise FirebaseAuthenticationError(f"An error occurred authenticating Firebase user => {err}") from err



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Firebase
   @package: firebase.core
      @file: file_processor.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
from abc import ABC
from firebase.domain.firebase_dto import FirebaseDto
from fnmatch import fnmatch
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.metaclass.classpath import AnyPath
from hspylib.core.preconditions import check_argument, check_not_none
from hspylib.core.tools.commons import syserr, sysout
from hspylib.modules.fetch.fetch import get, put
from requests.exceptions import HTTPError
from typing import List

import json
import logging as log
import os


class FileProcessor(ABC):
    """Utility class to upload / download B64-encoded files."""

    @staticmethod
    def upload_files(url: str, file_paths: List[AnyPath], glob_exp: str) -> int:
        """Upload files to URL.
        :param url: the URL to upload the files.
        :param file_paths: the file paths to be uploaded. File paths will be filtered by glob expressions.
        :param glob_exp: the GLOB expressions to filter the input files/folders.
        """
        data = []
        file_paths = list(map(os.path.expanduser, map(os.path.expandvars, file_paths)))
        for f_path in file_paths:
            if os.path.exists(f_path):
                if os.path.isfile(f_path):
                    sysout(f"%BLUE%Uploading file  {f_path:.<40} %NC%", end="")
                    dto = FileProcessor._read_and_encode(f_path)
                    data.append(dto)
                    sysout("[%GREEN%  OK  %NC%]")
                else:
                    sysout(f'%BLUE%Uploading files from directory "{f_path}" ')
                    all_files = next(os.walk(f_path))[2]
                    log.debug("\nGlob: %s \nFiles: %s", glob_exp, all_files)
                    for file in all_files:
                        filename = os.path.join(f_path, file)
                        if os.path.isfile(filename) and fnmatch(file, glob_exp or "*.*"):
                            sysout(f"%BLUE%Uploading file  {filename:.<40} %NC%", end="")
                            dto = FileProcessor._read_and_encode(filename)
                            data.append(dto)
                            sysout("[%GREEN%  OK  %NC%]")
            else:
                syserr(f'Input file "{f_path}" does not exist!')
        if data:
            payload = FileProcessor._create_request(data)
            response = put(url, payload)
            check_not_none(response)
            if response.status_code != HttpCode.OK:
                raise HTTPError(f"{response.status_code} - Unable to upload into={url} with json_string={payload}")
            paths = " \n  |- ".join([f.path for f in data])
            sysout(f"%EOL%%GREEN%File(s):%EOL%  |- {paths}%EOL%%EOL%Successfully uploaded to Firebase!%NC%")
            return len(data)

        syserr(f"No file has been uploaded: `{file_paths}`!")

        return 0

    @staticmethod
    def download_files(url: str, dest_dir: AnyPath) -> int:
        """Download files from URL.
        :param url: the URL to download the files.
        :param dest_dir: the destination directory.
        """
        check_argument(dest_dir and os.path.exists(dest_dir), "Unable find destination directory: {}", dest_dir)
        sysout(f'%BLUE%Downloading files from Firebase into "{dest_dir}" ...')
        response = get(url)
        check_not_none(response)
        check_not_none(response.body)
        if response.status_code != HttpCode.OK:
            raise HTTPError(f"{response.status_code} - Unable to download from={url} with response={response}")
        dto_list = FirebaseDto.from_json(response.body)
        if dto_list and len(dto_list) > 0:
            FileProcessor._decode_and_write(dest_dir, *dto_list)
            return len(dto_list)

        syserr(f"Database alias was not found in: {url} !")

        return 0

    @staticmethod
    def _read_and_encode(file_path: AnyPath) -> FirebaseDto:
        """Read and B64-encode a text file.
        :param file_path: the path to the encoding file.
        """
        return FirebaseDto(file_path).load().encode()

    @staticmethod
    def _decode_and_write(dest_dir: AnyPath, *data: FirebaseDto) -> None:
        """B64-decode and write entries to a text file.
        :param dest_dir: the destination directory.
        :param data: the file content to be written.
        """
        for entry in data:
            FirebaseDto.from_file(f"{dest_dir}/{os.path.basename(entry.path)}", entry.data).decode().save()
        paths = " \n  |- ".join([f.path for f in data])
        sysout(f'%EOL%%GREEN%File(s):%EOL%  |- {paths}%EOL%%EOL%Successfully downloaded into: "{dest_dir}"%NC%')

    @staticmethod
    def _create_request(entries: List[FirebaseDto]) -> str:
        """Create a Firebase request by converting a list of DTOs into JSON formatted string.
        :param entries: the file entries to be converted.
        """
        return f"[{','.join([json.dumps(e.__dict__) for e in entries])}]"
